import os
import warnings
from typing import Union
from pathlib import Path

from pydantic import field_validator

import torch
from torch.distributions import Normal
from torch.distributions.distribution import Distribution as TDistribution

from lume_model.models.prob_model_base import ProbModelBaseModel
from lume_model.models.torch_model import TorchModel


class NNEnsemble(ProbModelBaseModel):
    """LUME-model class for neural network ensembles.

    This class allows for the evaluation of multiple NN models as an ensemble.

    Args:
        models: List of one or more LUME-model neural network models (TorchModel instances).
    """

    models: list[TorchModel]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn("This class is still under development.")

    @field_validator("models", mode="before")
    def validate_torch_model_list(cls, v):
        if all(isinstance(m, (str, os.PathLike)) for m in v):
            for i, m in enumerate(v):
                fname = m.split("_model.pt")[0]
                if os.path.exists(m) and os.path.exists(f"{fname}.yml"):
                    # if it's a wrapper around TorchModel, might need a different class or a different way to load
                    v[i] = TorchModel(Path(f"{fname}.yml"))
                else:
                    raise OSError(
                        f"Both files, {m} and {fname}.yml, are required to load the models."
                    )

        if not all(isinstance(m, TorchModel) for m in v):
            raise TypeError("All models must be of type TorchModel.")

        return v

    def _get_predictions(
        self, input_dict: dict[str, float | torch.Tensor]
    ) -> dict[str, TDistribution]:
        """Get the predictions of the ensemble of models.
        This implements the abstract method from ProbModelBaseModel.

        Args:
            input_dict: Dictionary of input variable names to values.

        Returns:
            Dictionary of output variable names to distributions.
        """
        predictions = []
        for model in self.models:
            predictions.append(model.evaluate(input_dict))
        # Return a dictionary of output variable names to distributions
        return self._create_output_dict(predictions)

    def _create_output_dict(self, output_list: list) -> dict[str, TDistribution]:
        """Creates the output dictionary from the ensemble output.

        Args:
            output_list: List of output dictionaries.

        Returns:
            Dictionary of output variable names to distributions.
        """
        # Ensemble output is a list of dicts of output names to values
        # need to map them to a dict of output names to distributions
        ensemble_output_dict = {}

        for key in output_list[0]:  # for each named output
            output_tensor = torch.tensor([d[key].tolist() for d in output_list])
            ensemble_output_dict[key] = Normal(
                output_tensor.mean(axis=0),
                torch.sqrt(output_tensor.var(axis=0)),
            )
        return ensemble_output_dict

    @property
    def _tkwargs(self):
        """Returns the device and dtype for the model."""
        return {"device": self.device, "dtype": self.dtype}

    def dump(
        self,
        file: Union[str, os.PathLike],
        base_key: str = "",
        save_models: bool = True,
        save_jit: bool = False,
    ):
        """Dump the model to a file.

        Args:
            file: Path to the file to save the model.
            base_key: Base key for the model.
            save_models: Whether to save the models.
            save_jit: Whether to save the JIT models.
        """
        # Save each model in the ensemble
        mod_file = file.split(".yaml")[0].split(".yml")[0]
        for idx, model in enumerate(self.models):
            model.dump(
                f"{mod_file}_{idx}.yml",
                base_key=base_key,
                save_models=False,  # will be saved in the ensemble
                save_jit=False,
            )

        # Save the ensemble of models
        super().dump(file, base_key, save_models, save_jit)



import os
import yaml
from typing import Union

registered_models = []

# models requiring torch
try:
    from lume_model.models.torch_model import TorchModel
    from lume_model.models.torch_module import TorchModule
    from lume_model.models.ensemble import NNEnsemble
    from lume_model.models.gp_model import GPModel

    registered_models += [TorchModel, TorchModule, NNEnsemble, GPModel]
except ModuleNotFoundError:
    pass


def get_model(name: str):
    """Returns the LUME model class for the given name.

    Args:
        name: Name of LUME model class.

    Returns:
        LUME model class for the given name.
    """
    model_lookup = {m.__name__: m for m in registered_models}
    if name not in model_lookup.keys():
        raise KeyError(
            f"No model named {name}, available models are {list(model_lookup.keys())}"
        )
    return model_lookup[name]


def model_from_yaml(yaml_str: Union[str, os.PathLike]):
    """Creates LUME model from the given YAML formatted string or file path.

    Args:
        yaml_str: YAML formatted string or file path.

    Returns:
        Created LUME model.
    """
    if os.path.exists(yaml_str):
        with open(yaml_str) as f:
            config = yaml.safe_load(f.read())
    else:
        config = yaml.safe_load(yaml_str)
    model_class = get_model(config["model_class"])
    return model_class(yaml_str)



import pytest

try:
    from lume_model.models import TorchModel, TorchModule
    from lume_model.models import model_from_yaml
except ImportError:
    pass


@pytest.mark.parametrize(
    "filename,expected",
    [
        ("test_files/california_regression/torch_model.yml", TorchModel),
        ("test_files/california_regression/torch_module.yml", TorchModule),
    ],
)
def test_model_from_yaml(rootdir, filename, expected):
    model = model_from_yaml(f"{rootdir}/{filename}")
    assert isinstance(model, expected)



import os
import logging
from pydantic import field_validator

import torch
from torch.distributions import Distribution as TDistribution
from botorch.models import SingleTaskGP, MultiTaskGP, ModelListGP
from gpytorch.mlls import ExactMarginalLogLikelihood
from gpytorch.distributions import MultivariateNormal as GPMultivariateNormal
from botorch.models.transforms.input import ReversibleInputTransform
from botorch.models.transforms.outcome import OutcomeTransform
from linear_operator.utils.cholesky import psd_safe_cholesky
from linear_operator.operators import DiagLinearOperator

from lume_model.models.prob_model_base import (
    ProbModelBaseModel,
    TorchDistributionWrapper,
)


logger = logging.getLogger(__name__)


class GPModel(ProbModelBaseModel):
    """
    LUME-model class for GP models.
    This supports Botorch's SingleTask, MultiTask, and ModelListGP models.

    If input_transformers or output_transformers lists are passed, they will be applied sequentially to the
    inputs/outputs outside the underlying model, regardless of what the Botorch model's input_transform or
    outcome_transform attributes are set to (those transformations will still be handled internally by the Botorch
    model class). For ModelListGP, the passed input_transformers and output_transformers will be applied to all
    models in the list (outside the underlying models). If different transformers are needed for different models,
    the models should be instantiated separately using Botorch's input_transform and outcome_transform attributes
    before creating the model list.

    Args:
        model: A single task GPyTorch model or BoTorch model.
        input_transformers: List of input transformers to apply to the input data. They will be applied sequentially
            to the inputs. Optional, default is None.
        output_transformers: List of output transformers to apply to the output data. They will be applied sequentially
            to the outputs. Optional, default is None.
    """

    model: SingleTaskGP | MultiTaskGP | ModelListGP
    input_transformers: list[ReversibleInputTransform | torch.nn.Linear] | None = None
    output_transformers: (
        list[OutcomeTransform | ReversibleInputTransform | torch.nn.Linear] | None
    ) = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.input_transformers = (
            [] if self.input_transformers is None else self.input_transformers
        )
        self.output_transformers = (
            [] if self.output_transformers is None else self.output_transformers
        )

    @field_validator("model", mode="before")
    def validate_gp_model(cls, v):
        if isinstance(v, (str, os.PathLike)):
            if os.path.exists(v):
                v = torch.load(v, weights_only=False)
            else:
                raise OSError(f"File {v} is not found.")
        return v

    @field_validator("input_transformers", "output_transformers", mode="before")
    def validate_transformers(cls, v):
        if not isinstance(v, list):
            raise ValueError("Transformers must be passed as list.")
        loaded_transformers = []
        for t in v:
            if isinstance(t, (str, os.PathLike)):
                if os.path.exists(t):
                    t = torch.load(t, weights_only=False)
                else:
                    raise OSError(f"File {t} is not found.")
            loaded_transformers.append(t)
        v = loaded_transformers
        return v

    def get_input_size(self) -> int:
        """Get the dimensions of the input variables."""
        if isinstance(self.model, SingleTaskGP):
            num_inputs = self.model.train_inputs[0].shape[-1]
        elif isinstance(self.model, MultiTaskGP):
            num_inputs = self.model.train_inputs[0].shape[-1] - 1
        elif isinstance(self.model, ModelListGP):
            if isinstance(self.model.models[0], SingleTaskGP):
                num_inputs = self.model.models[0].train_inputs[0].shape[-1]
            elif isinstance(self.model.models[0], MultiTaskGP):
                num_inputs = self.model.models[0].train_inputs[0].shape[-1] - 1
        else:
            raise ValueError(
                "Model must be an instance of SingleTaskGP, MultiTaskGP or ModelListGP."
            )
        return num_inputs

    def get_output_size(self) -> int:
        """Get the dimensions of the output variables."""
        if isinstance(self.model, ModelListGP):
            num_outputs = sum(model.num_outputs for model in self.model.models)
        elif isinstance(self.model, SingleTaskGP) or isinstance(
            self.model, MultiTaskGP
        ):
            num_outputs = self.model.num_outputs
        else:
            raise ValueError(
                "Model must be an instance of SingleTaskGP, MultiTaskGP or ModelListGP."
            )
        return num_outputs

    @property
    def _tkwargs(self):
        """Returns the device and dtype for the model."""
        return {"device": self.device, "dtype": self.dtype}

    def likelihood(self):
        """Returns the likelihood of the model."""
        return self.model.likelihood

    def mll(self, x, y):
        """Returns the marginal log-likelihood value"""
        self.model.eval()
        mll = ExactMarginalLogLikelihood(self.model.likelihood, self.model)
        return mll(self.model(x), y).item()

    def _get_predictions(
        self,
        input_dict: dict[str, float | torch.Tensor],
        observation_noise: bool = False,
    ) -> dict[str, TDistribution]:
        """Get the predictions of the model.
        This implements the abstract method from ProbModelBaseModel.

        Args:
            input_dict: Dictionary of input variable names to values.

        Returns:
            Dictionary of output variable names to distributions.
        """
        # Create tensor from input_dict
        x = super()._create_tensor_from_dict(input_dict)
        # Transform the input
        if self.input_transformers is not None:
            x = self._transform_inputs(x)
        # Get the posterior distribution
        posterior = self._posterior(x, observation_noise=observation_noise)
        # Wrap the distribution in a torch distribution
        distribution = self._get_distribution(posterior)
        # Take mean and covariance of the distribution
        # posterior.mean preserves batch dim, while distribution.mean does not
        mean, covar = posterior.mean, distribution.covariance_matrix
        # Return a dictionary of output variable names to distributions
        return self._create_output_dict((mean, covar))

    def _posterior(self, x: torch.Tensor, observation_noise: bool = False):
        """Compute the posterior distribution.

        Args:
            x: Input tensor.

        Returns:
            Posterior object from the model.
        """
        self.model.eval()
        posterior = self.model.posterior(x, observation_noise=observation_noise)
        return posterior

    def _get_distribution(self, posterior) -> TDistribution:
        """Get the distribution from the posterior and checks that the covariance_matrix attribute exists.
        Args:
            posterior: Posterior object from the model.

        Returns:
            A torch distribution object.
        """
        if isinstance(posterior.distribution, TDistribution):
            d = posterior.distribution
        else:
            # Wrap the distribution in a torch distribution
            d = TorchDistributionWrapper(posterior.distribution)

        if not hasattr(d, "covariance_matrix"):
            raise ValueError(
                f"The posterior distribution {type(posterior.distribution)} does not have a covariance matrix attribute."
            )

        return d

    def _create_output_dict(
        self, output: tuple[torch.Tensor, torch.Tensor]
    ) -> dict[str, TDistribution]:
        """Returns outputs as dictionary of output names and their corresponding distributions.
        The returned distributions are constructed as torch multivariate normal distributions.
        At the moment, no other distribution types are supported.

        Args:
            output: Tuple containing mean and covariance of the output.

        Returns:
            Dictionary of output variable names to distributions. Distribution is a torch
            multivariate normal distribution.
        """
        output_distributions = {}
        mean, cov = output
        ss = mean.shape[1] if len(mean.shape) > 2 else mean.shape[0]  # sample size

        batch = mean.shape[0] if len(mean.shape) > 2 else None
        for i, name in enumerate(self.output_names):
            if batch is None:
                _mean = mean[:, i] if len(mean.shape) > 1 else mean
                _cov = torch.zeros(ss, ss, **self._tkwargs)
                _cov[:, :ss] = cov[i * ss : (i + 1) * ss, i * ss : (i + 1) * ss]
            else:
                _mean = mean[:, :, i]
                _cov = torch.zeros(batch, ss, ss, **self._tkwargs)
                _cov[:, :ss, :ss] = cov[:, i * ss : (i + 1) * ss, i * ss : (i + 1) * ss]

            # Check that the covariance matrix is positive definite
            _cov = self._check_covariance_matrix(_cov)

            if self.output_transformers is not None:
                # TODO: make this more robust?
                # If we have two outputs, but transformer has length 1 (e.g. multitask),
                # we should apply the same transform to both outputs
                _mean = self._transform_mean(_mean, i)
                _cov = self._transform_covar(_cov, i)

            output_distributions[name] = GPMultivariateNormal(_mean, _cov)

        return output_distributions

    def _transform_inputs(self, input_tensor: torch.Tensor) -> torch.Tensor:
        """Applies transformations to the inputs.

        Args:
            input_tensor: Ordered input tensor to be passed to the transformers.

        Returns:
            Tensor of transformed inputs to be passed to the model.
        """
        for transformer in self.input_transformers:
            if isinstance(transformer, ReversibleInputTransform):
                input_tensor = transformer.transform(input_tensor)
            else:
                input_tensor = transformer(input_tensor)
        return input_tensor

    def _transform_mean(self, mean: torch.Tensor, i) -> torch.Tensor:
        """(Un-)Transforms the model output mean.

        Args:
            mean: Output mean tensor from the model.

        Returns:
            (Un-)Transformed output mean tensor.
        """
        for transformer in self.output_transformers:
            if isinstance(transformer, ReversibleInputTransform):
                try:
                    scale_fac = transformer.coefficient[i]
                    offset = transformer.offset[i]
                except IndexError:
                    # If the transformer has only one coefficient, use it for all outputs
                    # This is needed in the case of multitask models
                    scale_fac = transformer.coefficient[0]
                    offset = transformer.offset[0]
                mean = offset + scale_fac * mean
            elif isinstance(transformer, OutcomeTransform):
                try:
                    scale_fac = transformer.stdvs.squeeze(0)[i]
                    offset = transformer.means.squeeze(0)[i]
                except IndexError:
                    # If the transformer has only one coefficient, use it for all outputs
                    scale_fac = transformer.stdvs.squeeze(0)[0]
                    offset = transformer.means.squeeze(0)[0]
                mean = offset + scale_fac * mean
            else:
                raise NotImplementedError(
                    f"Output transformer {type(transformer)} is not supported."
                )
        return mean

    def _transform_covar(self, cov: torch.Tensor, i: int) -> torch.Tensor:
        """(Un-)Transforms the model output covariance matrix.

        Args:
            cov: Output covariance matrix tensor from the model.
            i: Index of the output variable.

        Returns:
            (Un-)Transformed output covariance matrix tensor.
        """
        for transformer in self.output_transformers:
            if isinstance(transformer, ReversibleInputTransform):
                try:
                    scale_fac = transformer.coefficient[i]
                except IndexError:
                    # If the transformer has only one coefficient, use it for all outputs
                    scale_fac = transformer.coefficient[0]
                scale_fac = scale_fac.expand(cov.shape[:-1])
                scale_mat = DiagLinearOperator(scale_fac)
                cov = scale_mat @ cov @ scale_mat
            elif isinstance(transformer, OutcomeTransform):
                try:
                    scale_fac = transformer.stdvs.squeeze(0)[i]
                except IndexError:
                    # If the transformer has only one coefficient, use it for all outputs
                    scale_fac = transformer.stdvs.squeeze(0)[0]
                scale_fac = scale_fac.expand(cov.shape[:-1])
                scale_mat = DiagLinearOperator(scale_fac)
                cov = scale_mat @ cov @ scale_mat
            else:
                raise NotImplementedError(
                    f"Output transformer {type(transformer)} is not supported."
                )
        return cov

    def _check_covariance_matrix(self, cov: torch.Tensor) -> torch.Tensor:
        """Checks that the covariance matrix is positive definite, and adds jitter if not."""
        try:
            torch.linalg.cholesky(cov)
        except torch._C._LinAlgError:
            lm = psd_safe_cholesky(cov)  # determines jitter iteratively
            cov = lm @ lm.transpose(-1, -2)

        return cov



from typing import Union, Any, Tuple
from abc import abstractmethod

from pydantic import model_validator
import torch
from torch.distributions import Distribution as TDistribution

from lume_model.variables import DistributionVariable
from lume_model.models.utils import InputDictModel, format_inputs, itemize_dict
from lume_model.base import LUMEBaseModel


class ProbModelBaseModel(LUMEBaseModel):  # TODO: brainstorm a better name
    """Abstract base class for probabilistic models.

    This class provides a common interface for probabilistic models. All subclasses need to
     implement a `_get_predictions` method that accepts a dictionary input, and returns a
     dictionary of output names to _distributions_. The distributions should be instances of
    `torch.distributions.Distribution`.

    Attributes:
        output_variables: List of output variables, which should be of DistributionVariable type.
        device: Device on which the model will be evaluated. Defaults to "cpu".
        precision: Precision of the model, either "double" or "single". Defaults to "double".
    """

    output_variables: list[DistributionVariable]
    device: Union[torch.device, str] = "cpu"
    precision: str = "double"

    @model_validator(mode="before")
    def validate_output_variables(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate output variables as DistributionVariable."""
        for variable in values["output_variables"]:
            if not isinstance(variable, DistributionVariable):
                raise ValueError(
                    "Output variables must be of type DistributionVariable."
                )
        return values

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # No range validation for probabilistic models currently implemented
        # self.input_validation_config = {x: "none" for x in self.input_names}

    @property
    def dtype(
        self,
    ):
        """Returns the data type for the model."""
        if self.precision == "double":
            return torch.double
        elif self.precision == "single":
            return torch.float
        else:
            raise ValueError(
                f"Unknown precision {self.precision}, "
                f"expected one of ['double', 'single']."
            )

    @staticmethod
    def _create_tensor_from_dict(
        d: dict[str, Union[float, torch.Tensor]],
    ) -> torch.Tensor:
        """Create a 2D tensor from a dictionary of floats and tensors.

        Args:
            d: Dictionary of floats or tensors.

        Returns:
            A Torch Tensor."""
        tensors = []

        for key, value in d.items():
            if isinstance(value, float):
                tensors.append(torch.tensor([value]))
            elif isinstance(value, torch.Tensor):
                tensors.append(value)
            else:
                raise ValueError(
                    f"Value for key '{key}' must be either a float or a torch tensor."
                )

        if all(isinstance(value, float) for value in d.values()):
            # All values are floats
            return torch.stack(tensors, dim=1)
        elif all(isinstance(tensor, torch.Tensor) for tensor in tensors):
            lengths = [tensor.size(0) for tensor in tensors]
            if len(set(lengths)) != 1:
                raise ValueError("All tensors must have the same length.")
            dim = tensors[0].dim()
            # Stack tensors into a multidimensional tensor
            return torch.stack(tensors, dim=dim)
        else:
            raise ValueError(
                "All values must be either floats or tensors, and all tensors must have the same length."
            )

    @abstractmethod
    def _get_predictions(
        self, input_dict: dict[str, float | torch.Tensor], **kwargs
    ) -> dict[str, TDistribution]:
        """Get predictions from the model.

        Args:
             input_dict: Dictionary of input variable names to values. Values can be floats or
                `n` or `b x n` (batch mode) torch tensors.
        Returns:
            A dictionary of output variable names to distributions.
        """
        pass

    def _evaluate(
        self, input_dict: dict[str, Union[float, torch.Tensor]], **kwargs
    ) -> dict[str, TDistribution]:
        """Evaluate the model.

        Args:
            input_dict: Dictionary of input variable names to values. Values can be floats or
                `n` or `b x n` (batch mode) torch tensors.

        Returns:
            A dictionary of output variable names to distributions.
        """
        # Evaluate and get mean and variance for each output
        output_dict = self._get_predictions(input_dict, **kwargs)
        # Split multi-dimensional output into separate distributions and
        # return output dictionary
        return output_dict

    def input_validation(self, input_dict: dict[str, Union[float, torch.Tensor]]):
        """Validates input dictionary before evaluation.

        Args:
            input_dict: Input dictionary to validate.

        Returns:
            Validated input dictionary.
        """
        # validate input type (ints only are cast to floats for scalars)
        validated_input = InputDictModel(input_dict=input_dict).input_dict
        # format inputs as tensors w/o changing the dtype
        formatted_inputs = format_inputs(validated_input)
        # itemize inputs for validation
        itemized_inputs = itemize_dict(formatted_inputs)

        for ele in itemized_inputs:
            # validate values that were in the torch tensor
            # any ints in the torch tensor will be cast to floats by Pydantic
            # but others will be caught, e.g. booleans
            ele = InputDictModel(input_dict=ele).input_dict
            # validate each value based on its var class and config
            super().input_validation(ele)

        # return the validated input dict for consistency w/ casting ints to floats
        if any([isinstance(value, torch.Tensor) for value in validated_input.values()]):
            validated_input = {
                k: v.to(**self._tkwargs).squeeze(-1) for k, v in validated_input.items()
            }

        return validated_input

    def output_validation(self, output_dict: dict[str, TDistribution]):
        """Itemizes tensors before performing output validation."""
        itemized_outputs = itemize_dict(output_dict)
        for ele in itemized_outputs:
            super().output_validation(ele)


class TorchDistributionWrapper(TDistribution):
    """Wraps any distribution to provide a torch.distributions-like interface."""

    def __init__(self, custom_dist):
        """
        Args:
            custom_dist: An instance of a custom distribution with methods:
                          mean, variance, log_prob, sample, and rsample.
        """
        super().__init__()
        self.custom_dist = custom_dist
        self.device = torch.device("cpu")
        self.dtype = torch.double

    @property
    def _tkwargs(self):
        return {"device": self.device, "dtype": self.dtype}

    @property
    def mean(self) -> torch.Tensor:
        """Return the mean of the custom distribution."""
        attribute_names = ["mean"]
        result, _ = self._get_attr(attribute_names)
        return result

    @property
    def variance(self) -> torch.Tensor:
        """Return the variance of the custom distribution."""
        attribute_names = ["variance", "var", "cov", "covariance", "covariance_matrix"]
        result, attr_name = self._get_attr(attribute_names)

        if attr_name in ["cov", "covariance", "covariance_matrix"]:
            return torch.diagonal(torch.tensor(result))

        return result

    @property
    def covariance_matrix(self) -> torch.Tensor:
        """Return the covariance matrix of the custom distribution."""
        attribute_names = ["covariance_matrix", "cov", "covariance"]
        result, _ = self._get_attr(attribute_names)
        return result

    def confidence_region(self) -> Tuple[torch.tensor, torch.tensor]:
        """
        Adapted from gpytorch.distributions.multivariate_normal
        Returns 2 standard deviations above and below the mean.

        Returns:
            Pair of tensors of size `... x N`, where N is the
            dimensionality of the random variable. The first (second) Tensor is the
            lower (upper) end of the confidence region.
        """
        try:
            stddev = self.variance.sqrt()
            std2 = stddev.mul_(2)
            mean = self.mean
            return mean.sub(std2), mean.add(std2)
        except AttributeError:
            raise AttributeError("The distribution does not have a variance attribute.")

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        """Compute the log probability for a given value."""
        attribute_names = ["log_prob", "log_likelihood", "logpdf"]
        result, _ = self._get_attr(attribute_names, value)
        return result

    # TODO: check fn signature
    def rsample(self, sample_shape: torch.Size()) -> torch.Tensor:
        """Generate reparameterized samples from the custom distribution."""
        # Fallback to sample if rsample is not implemented
        attribute_names = ["rsample", "sample", "rvs"]
        result, _ = self._get_attr(attribute_names, sample_shape)
        return result

    def sample(self, sample_shape: torch.Size()) -> torch.Tensor:
        """Generate samples from the custom distribution (non-differentiable if using sample)."""
        attribute_names = ["sample", "rvs"]
        # Assume non-torch.Distribution takes an integer sample_shape
        sample_shape = sample_shape.numel()
        result, _ = self._get_attr(attribute_names, sample_shape)
        return result

    def __repr__(self):
        return f"TorchDistributionWrapper({self.custom_dist})"

    def _get_attr(self, attribute_names, value=None):
        """Get the first attribute that is found in the distribution."""
        for attr_name in attribute_names:
            attr_value = getattr(self.custom_dist, attr_name, None)
            if attr_value is not None:
                if callable(attr_value):
                    result = attr_value(value) if value is not None else attr_value()
                else:
                    result = attr_value

                return torch.tensor(result, **self._tkwargs), attr_name

        raise AttributeError(
            f"None of the attributes {attribute_names} found in the distribution."
        )



import os
import random
from typing import Any, Union
from copy import deepcopy

import pytest

try:
    import torch
    from botorch.models.transforms.input import AffineInputTransform
    from lume_model.models import TorchModel
    from lume_model.variables import ScalarVariable

    torch.manual_seed(42)
except ImportError:
    pass

random.seed(42)


def assert_model_equality(m1: TorchModel, m2: TorchModel):
    assert m1.input_variables == m2.input_variables
    assert m1.output_variables == m2.output_variables
    assert str(m1.model.state_dict()) == str(m2.model.state_dict())
    for attr in ["input_transformers", "output_transformers"]:
        m1_transformers, m2_transformers = getattr(m1, attr), getattr(m2, attr)
        assert len(m1_transformers) == len(m2_transformers)
        for i, t in enumerate(m1_transformers):
            if isinstance(t, torch.nn.Module):
                assert isinstance(m2_transformers[i], torch.nn.Module)
                assert str(t.state_dict()) == str(m2_transformers[i].state_dict())
    assert m1.output_format == m2.output_format
    assert m1.device == m2.device
    assert m1.fixed_model == m2.fixed_model


class TestTorchModel:
    def test_model_from_objects(
        self,
        california_model_info: dict[str, str],
        california_model_kwargs: dict[str, Union[list, dict, str]],
        california_variables: tuple[list[ScalarVariable], list[ScalarVariable]],
        california_transformers: tuple[list, list],
        california_model,
    ):
        input_variables, output_variables = california_variables
        input_transformer, output_transformer = california_transformers

        assert isinstance(california_model, TorchModel)
        assert california_model.input_names == california_model_info["model_in_list"]
        assert california_model.output_names == california_model_info["model_out_list"]
        assert california_model.input_variables == input_variables
        assert california_model.output_variables == output_variables
        assert california_model.input_transformers == [input_transformer]
        assert california_model.output_transformers == [output_transformer]

    def test_model_from_yaml(self, rootdir: str, california_model):
        file = os.path.join(
            rootdir, "test_files", "california_regression", "torch_model.yml"
        )
        yaml_model = TorchModel.from_file(file)

        assert_model_equality(yaml_model, california_model)

    def test_model_as_yaml(self, rootdir: str, california_model):
        filename = "test_torch_model"
        file = f"{filename}.yml"
        california_model.dump(file)
        yaml_model = TorchModel(file)
        assert_model_equality(yaml_model, california_model)
        os.remove(file)
        os.remove(f"{filename}_model.pt")
        os.remove(f"{filename}_input_transformers_0.pt")
        os.remove(f"{filename}_output_transformers_0.pt")

    def test_input_validation(self, california_test_input_dict: dict, california_model):
        california_model.input_validation(california_test_input_dict)

    def test_output_validation(self, california_model):
        output_dict = {"MedHouseVal": torch.tensor([5.0, 3.1])}
        california_model.output_validation(output_dict)

    def test_precision(self, california_model):
        assert california_model.precision == "double"
        assert california_model.dtype == torch.double
        california_model.precision = "single"
        assert california_model.dtype == torch.float
        # set back to double
        california_model.precision = "double"

    def test_model_evaluate_single_sample(
        self, california_test_input_dict: dict, california_model
    ):
        results = california_model.evaluate(california_test_input_dict)

        assert isinstance(results["MedHouseVal"], torch.Tensor)
        assert torch.isclose(
            results["MedHouseVal"],
            torch.tensor(4.063651, dtype=results["MedHouseVal"].dtype),
        )

    def test_model_evaluate_n_samples(
        self, california_test_input_tensor, california_model
    ):
        test_dict = {
            key: california_test_input_tensor[:, idx]
            for idx, key in enumerate(california_model.input_names)
        }
        results = california_model.evaluate(test_dict)
        target_tensor = torch.tensor(
            [4.063651, 2.7774928, 2.792812], dtype=results["MedHouseVal"].dtype
        )

        assert torch.all(torch.isclose(results["MedHouseVal"], target_tensor))

    def test_model_evaluate_batch_n_samples(
        self,
        california_test_input_tensor,
        california_model,
    ):
        # model should be able to handle input of shape [n_batch, n_samples, n_dim]
        input_dict = {
            key: california_test_input_tensor[:, idx]
            .unsqueeze(-1)
            .unsqueeze(1)
            .repeat((1, 3, 1))
            for idx, key in enumerate(california_model.input_names)
        }
        results = california_model.evaluate(input_dict)

        # output shape should be [n_batch, n_samples]
        assert tuple(results["MedHouseVal"].shape) == (3, 3)

    def test_model_evaluate_raw(
        self,
        california_test_input_dict: dict,
        california_model_kwargs: dict[str, Union[list, dict, str]],
    ):
        kwargs = deepcopy(california_model_kwargs)
        kwargs["output_format"] = "raw"
        california_model = TorchModel(**kwargs)
        float_dict = {
            key: value.item() for key, value in california_test_input_dict.items()
        }
        results = california_model.evaluate(float_dict)

        assert isinstance(results["MedHouseVal"], float)
        assert results["MedHouseVal"] == pytest.approx(4.063651)

    def test_model_evaluate_shuffled_input(
        self, california_test_input_dict: dict, california_model
    ):
        shuffled_input = deepcopy(california_test_input_dict)
        item_list = list(shuffled_input.items())
        random.shuffle(item_list)
        shuffled_input = dict(item_list)
        results = california_model.evaluate(shuffled_input)

        assert isinstance(results["MedHouseVal"], torch.Tensor)
        assert torch.isclose(
            results["MedHouseVal"],
            torch.tensor(4.063651, dtype=results["MedHouseVal"].dtype),
        )

    @pytest.mark.parametrize(
        "test_idx,expected", [(0, 4.063651), (1, 2.7774928), (2, 2.792812)]
    )
    def test_model_evaluate_different_values(
        self,
        test_idx: int,
        expected: float,
        california_test_input_tensor,
        california_model,
    ):
        input_dict = {
            key: california_test_input_tensor[test_idx][idx]
            for idx, key in enumerate(california_model.input_names)
        }
        results = california_model.evaluate(input_dict)

        assert results["MedHouseVal"].item() == pytest.approx(expected)

    def test_model_evaluate_with_no_output_transformers(
        self,
        california_test_input_dict: dict,
        california_model_kwargs: dict[str, Union[list, dict, str]],
    ):
        kwargs = deepcopy(california_model_kwargs)
        kwargs["output_transformers"] = []
        model = TorchModel(**kwargs)
        results = model.evaluate(california_test_input_dict)

        assert torch.isclose(
            results["MedHouseVal"],
            torch.tensor(1.8523695, dtype=results["MedHouseVal"].dtype),
        )

    def test_differentiability(
        self,
        california_test_input_dict: dict,
        california_model_kwargs: dict[str, Any],
    ):
        kwargs = deepcopy(california_model_kwargs)
        kwargs["model"].train().requires_grad_(True)
        model = TorchModel(**kwargs, fixed_model=False)
        parameters_with_requires_grad = []
        for name, param in model.model.named_parameters():
            if param.requires_grad:
                parameters_with_requires_grad.append(name)
        criterion = torch.nn.MSELoss()

        assert model.model.training
        assert len(parameters_with_requires_grad) == 8
        output_dict = model.evaluate(california_test_input_dict)
        outputs = torch.stack([v for k, v in output_dict.items()])
        loss = criterion(outputs, torch.zeros(outputs.shape, dtype=outputs.dtype))
        loss.backward()

    def test_update_input_variables_to_transformer(self, california_model):
        model = deepcopy(california_model)
        input_variables = model.input_variables

        def get_x_limits(v):
            x_limits = {
                "min": torch.tensor(
                    [var.value_range[0] for var in v], dtype=california_model.dtype
                ),
                "max": torch.tensor(
                    [var.value_range[1] for var in v], dtype=california_model.dtype
                ),
                "default": torch.tensor(
                    [var.default_value for var in v], dtype=california_model.dtype
                ),
            }
            return x_limits

        x_lim = get_x_limits(input_variables)
        x_lim_nn = {key: model._transform_inputs(x_lim[key]) for key in x_lim.keys()}
        # add new transformer
        d = len(model.input_variables)
        new_transformer = AffineInputTransform(
            d=d, offset=torch.rand(d), coefficient=1.0 + torch.rand(d)
        )
        model.insert_input_transformer(
            new_transformer, loc=len(model.input_transformers)
        )
        updated_input_variables = model.update_input_variables_to_transformer(
            len(model.input_transformers) - 1
        )
        model.input_variables = updated_input_variables
        # compute new NN limits
        x_lim_updated = get_x_limits(updated_input_variables)
        x_lim_nn_updated = {
            key: model._transform_inputs(x_lim_updated[key])
            for key in x_lim_updated.keys()
        }

        for key in x_lim_nn.keys():
            assert torch.all(
                torch.isclose(x_lim_nn[key], x_lim_nn_updated[key], atol=1e-6)
            )

    def test_jit_serialization(self, california_model):
        filename = "test_torch_model"
        file = f"{filename}.yml"

        # Test saving with save_jit=True
        california_model.dump(file, save_jit=True)
        assert os.path.exists(file)
        assert os.path.exists(f"{filename}_model.jit")

        # Test loading the JIT model
        try:
            loaded_model = torch.jit.load(f"{filename}_model.jit")
            assert loaded_model is not None
        except Exception as e:
            pytest.fail(f"Failed to load JIT model: {e}")

        os.remove(file)
        os.remove(f"{filename}_model.pt")
        os.remove(f"{filename}_model.jit")
        os.remove(f"{filename}_input_transformers_0.pt")
        os.remove(f"{filename}_output_transformers_0.pt")



import os
import json
import logging
from abc import ABC, abstractmethod
from typing import Any, Callable, Optional, Union
from types import FunctionType, MethodType
from io import TextIOWrapper

import yaml
import numpy as np
from pydantic import BaseModel, ConfigDict, field_validator
import torch  # TODO: for torch.Tensor type hinting, but may need to make more general in mlflow class

from lume_model.variables import ScalarVariable, get_variable, ConfigEnum
from lume_model.utils import (
    try_import_module,
    verify_unique_variable_names,
    serialize_variables,
    deserialize_variables,
    variables_from_dict,
    replace_relative_paths,
)
from lume_model.mlflow_utils import register_model

logger = logging.getLogger(__name__)

JSON_ENCODERS = {
    # function/method type distinguished for class members and not recognized as callables
    FunctionType: lambda x: f"{x.__module__}.{x.__qualname__}",
    MethodType: lambda x: f"{x.__module__}.{x.__qualname__}",
    Callable: lambda x: f"{x.__module__}.{x.__qualname__}",
    type: lambda x: f"{x.__module__}.{x.__name__}",
    np.ndarray: lambda x: x.tolist(),
    np.int64: lambda x: int(x),
    np.float64: lambda x: float(x),
}


def process_torch_module(
    module,
    base_key: str = "",
    key: str = "",
    file_prefix: Union[str, os.PathLike] = "",
    save_modules: bool = True,
    save_jit: bool = False,
):
    """Optionally saves the given torch module to file and returns the filename.

    Args:
        base_key: Base key at this stage of serialization.
        key: Key corresponding to the torch module.
        module: The torch module to process.
        file_prefix: Prefix for generated filenames.
        save_modules: Determines whether torch modules are saved to file.
        save_jit: Determines whether the model gets saved as TorchScript.

    Returns:
        Filename under which the torch module is (or would be) saved.
    """
    torch = try_import_module("torch")
    filepath_prefix, filename_prefix = os.path.split(file_prefix)
    prefixes = [ele for ele in [filename_prefix, base_key] if not ele == ""]
    filename = "{}.pt".format(key)
    jit_filename = "{}.jit".format(key)
    if prefixes:
        filename = "_".join((*prefixes, filename))
        jit_filename = "_".join((*prefixes, jit_filename))
    if save_modules:
        filepath = os.path.join(filepath_prefix, filename)
        torch.save(module, filepath)
    if save_jit:
        filepath = os.path.join(filepath_prefix, jit_filename)
        try:
            scripted_model = torch.jit.script(module)
            torch.jit.save(scripted_model, filepath)
        except Exception as e:
            logger.warning(
                "Saving as JIT through scripting has only been evaluated "
                "for NN models that don't depend on BoTorch modules."
            )
            logger.error(f"Failed to script the model: {e}")
            raise e
    return jit_filename if save_jit else filename


def recursive_serialize(
    v: dict[str, Any],
    base_key: str = "",
    file_prefix: Union[str, os.PathLike] = "",
    save_models: bool = True,
    save_jit: bool = False,
):
    """Recursively performs custom serialization for the given object.

    Args:
        v: Object to serialize.
        base_key: Base key at this stage of serialization.
        file_prefix: Prefix for generated filenames.
        save_models: Determines whether models are saved to file.
        save_jit: Determines whether the model is saved as TorchScript.

    Returns:
        Serialized object.
    """
    # try to import modules for LUMEBaseModel child classes
    torch = try_import_module("torch")
    # serialize
    v = serialize_variables(v)
    for key, value in v.items():
        if isinstance(value, dict):
            v[key] = recursive_serialize(value, key)
        elif isinstance(value, list) and all(isinstance(ele, dict) for ele in value):
            # e.g. NN ensemble
            v[key] = [
                recursive_serialize(value[i], f"{base_key}{i}", file_prefix)
                for i in range(len(value))
            ]
            # For NN ensembles, we want v[key] to be a list of the filenames corresponding to each
            # model in the ensemble and not the serialized dict of each
            # NOTE: If this clause is reached for other models, we may need to do this differently
            v[key] = [v[key][i]["model"] for i in range(len(value))]
        elif torch is not None and isinstance(value, torch.nn.Module):
            v[key] = process_torch_module(
                value, base_key, key, file_prefix, save_models, save_jit
            )
        elif (
            isinstance(value, list)
            and torch is not None
            and any(isinstance(ele, torch.nn.Module) for ele in value)
        ):
            # List of transformers
            v[key] = [
                process_torch_module(
                    value[i], base_key, f"{key}_{i}", file_prefix, save_models, False
                )
                for i in range(len(value))
            ]
        else:
            for _type, func in JSON_ENCODERS.items():
                if isinstance(value, _type):
                    v[key] = func(value)
        # check to make sure object has been serialized, if not use a generic serializer
        try:
            json.dumps(v[key])
        except (TypeError, OverflowError):
            v[key] = f"{v[key].__module__}.{v[key].__class__.__qualname__}"

    return v


def recursive_deserialize(v):
    """Recursively performs custom deserialization for the given object.

    Args:
        v: Object to deserialize.

    Returns:
        Deserialized object.
    """
    # deserialize
    v = deserialize_variables(v)
    for key, value in v.items():
        if isinstance(value, dict):
            v[key] = recursive_deserialize(value)
    return v


def json_dumps(
    v,
    *,
    base_key="",
    file_prefix: Union[str, os.PathLike] = "",
    save_models: bool = True,
    save_jit: bool = False,
):
    """Serializes variables before dumping with json.

    Args:
        v: Object to dump.
        base_key: Base key for serialization.
        file_prefix: Prefix for generated filenames.
        save_models: Determines whether models are saved to file.
        save_jit: Determines whether the model is saved as TorchScript.

    Returns:
        JSON formatted string.
    """
    v = recursive_serialize(
        v.model_dump(), base_key, file_prefix, save_models, save_jit
    )
    v = json.dumps(v)
    return v


def json_loads(v):
    """Loads JSON formatted string and recursively deserializes the result.

    Args:
        v: JSON formatted string to load.

    Returns:
        Deserialized object.
    """
    v = json.loads(v)
    v = recursive_deserialize(v)
    return v


def parse_config(
    config: Union[dict, str, TextIOWrapper, os.PathLike],
    model_fields: dict = None,
) -> dict:
    """Parses model configuration and returns keyword arguments for model constructor.

    Args:
        config: Model configuration as dictionary, YAML or JSON formatted string, file or file path.
        model_fields: Fields expected by the model (required for replacing relative paths).

    Returns:
        Configuration as keyword arguments for model constructor.
    """
    config_file = None
    if isinstance(config, dict):
        d = config
    else:
        if isinstance(config, TextIOWrapper):
            yaml_str = config.read()
            config_file = os.path.abspath(config.name)
        elif isinstance(config, (str, os.PathLike)) and os.path.exists(config):
            with open(config) as f:
                yaml_str = f.read()
            config_file = os.path.abspath(config)
        else:
            yaml_str = config
        d = recursive_deserialize(yaml.safe_load(yaml_str))
    if config_file is not None:
        config_dir = os.path.dirname(os.path.realpath(config_file))
        d = replace_relative_paths(d, model_fields, config_dir)
    return model_kwargs_from_dict(d)


def model_kwargs_from_dict(config: dict) -> dict:
    """Processes model configuration and returns the corresponding keyword arguments for model constructor.

    Args:
        config: Model configuration.

    Returns:
        Configuration as keyword arguments for model constructor.
    """
    config = deserialize_variables(config)
    if all(key in config.keys() for key in ["input_variables", "output_variables"]):
        config["input_variables"], config["output_variables"] = variables_from_dict(
            config
        )
    config.pop("model_class", None)
    return config


class LUMEBaseModel(BaseModel, ABC):
    """Abstract base class for models using lume-model variables.

    Inheriting classes must define the evaluate method and variable names must be unique (respectively).
    Models build using this framework will be compatible with the lume-epics EPICS server and associated tools.

    Attributes:
        input_variables: List defining the input variables and their order.
        output_variables: List defining the output variables and their order.
        input_validation_config: Determines the behavior during input validation by specifying the validation
          config for each input variable: {var_name: value}. Value can be "warn", "error", or None.
        output_validation_config: Determines the behavior during output validation by specifying the validation
          config for each output variable: {var_name: value}. Value can be "warn", "error", or None.
    """

    input_variables: list[ScalarVariable]
    output_variables: list[ScalarVariable]
    input_validation_config: Optional[dict[str, ConfigEnum]] = None
    output_validation_config: Optional[dict[str, ConfigEnum]] = None

    model_config = ConfigDict(arbitrary_types_allowed=True, validate_assignment=True)

    @field_validator("input_variables", "output_variables", mode="before")
    def validate_input_variables(cls, value):
        new_value = []
        if isinstance(value, dict):
            for name, val in value.items():
                if isinstance(val, dict):
                    variable_class = get_variable(val["variable_class"])
                    new_value.append(variable_class(name=name, **val))
                elif isinstance(val, ScalarVariable):
                    new_value.append(val)
                else:
                    raise TypeError(f"type {type(val)} not supported")
        elif isinstance(value, list):
            new_value = value
        return new_value

    def __init__(self, *args, **kwargs):
        """Initializes LUMEBaseModel.

        Args:
            *args: Accepts a single argument which is the model configuration as dictionary, YAML or JSON
              formatted string or file path.
            **kwargs: See class attributes.
        """
        if len(args) == 1:
            if len(kwargs) > 0:
                raise ValueError(
                    "Cannot specify YAML string and keyword arguments for LUMEBaseModel init."
                )
            super().__init__(**parse_config(args[0], type(self).model_fields))
        elif len(args) > 1:
            raise ValueError(
                "Arguments to LUMEBaseModel must be either a single YAML string "
                "or keyword arguments passed directly to pydantic."
            )
        else:
            super().__init__(**kwargs)

    @field_validator("input_variables", "output_variables")
    def unique_variable_names(cls, value):
        verify_unique_variable_names(value)
        return value

    @property
    def input_names(self) -> list[str]:
        return [var.name for var in self.input_variables]

    @property
    def output_names(self) -> list[str]:
        return [var.name for var in self.output_variables]

    @property
    def default_input_validation_config(self) -> dict[str, ConfigEnum]:
        """Determines default behavior during input validation (if input_validation_config is None)."""
        return {var.name: var.default_validation_config for var in self.input_variables}

    @property
    def default_output_validation_config(self) -> dict[str, ConfigEnum]:
        """Determines default behavior during output validation (if output_validation_config is None)."""
        return {
            var.name: var.default_validation_config for var in self.output_variables
        }

    def evaluate(self, input_dict: dict[str, Any], **kwargs) -> dict[str, Any]:
        """Main evaluation function, child classes must implement the _evaluate method."""
        validated_input_dict = self.input_validation(input_dict)
        output_dict = self._evaluate(validated_input_dict, **kwargs)
        self.output_validation(output_dict)
        return output_dict

    @abstractmethod
    def _evaluate(self, input_dict: dict[str, Any], **kwargs) -> dict[str, Any]:
        pass

    def input_validation(self, input_dict: dict[str, Any]) -> dict[str, Any]:
        for name, value in input_dict.items():
            _config = (
                None
                if self.input_validation_config is None
                else self.input_validation_config.get(name)
            )
            var = self.input_variables[self.input_names.index(name)]
            var.validate_value(value, config=_config)
        return input_dict

    def output_validation(self, output_dict: dict[str, Any]) -> dict[str, Any]:
        for name, value in output_dict.items():
            _config = (
                None
                if self.output_validation_config is None
                else self.output_validation_config.get(name)
            )
            var = self.output_variables[self.output_names.index(name)]
            var.validate_value(value, config=_config)
        return output_dict

    def to_json(self, **kwargs) -> str:
        return json_dumps(self, **kwargs)

    def model_dump(self, **kwargs) -> dict[str, Any]:
        config = super().model_dump(**kwargs)
        config["input_variables"] = [var.model_dump() for var in self.input_variables]
        config["output_variables"] = [var.model_dump() for var in self.output_variables]
        return {"model_class": self.__class__.__name__} | config

    def json(self, **kwargs) -> str:
        result = self.to_json(**kwargs)
        config = json.loads(result)
        return json.dumps(config)

    def yaml(
        self,
        base_key: str = "",
        file_prefix: str = "",
        save_models: bool = False,
        save_jit: bool = False,
    ) -> str:
        """Serializes the object and returns a YAML formatted string defining the model.

        Args:
            base_key: Base key for serialization.
            file_prefix: Prefix for generated filenames.
            save_models: Determines whether models are saved to file.
            save_jit: Determines whether the model is saved as TorchScript
        Returns:
            YAML formatted string defining the model.
        """
        output = json.loads(
            self.to_json(
                base_key=base_key,
                file_prefix=file_prefix,
                save_models=save_models,
                save_jit=save_jit,
            )
        )
        s = yaml.dump(output, default_flow_style=None, sort_keys=False)
        return s

    def dump(
        self,
        file: Union[str, os.PathLike],
        base_key: str = "",
        save_models: bool = True,
        save_jit: bool = False,
    ):
        """Returns and optionally saves YAML formatted string defining the model.

        Args:
            file: File path to which the YAML formatted string and corresponding files are saved.
            base_key: Base key for serialization.
            save_models: Determines whether models are saved to file.
            save_jit: Determines whether the model is saved as TorchScript.
        """
        file_prefix = os.path.splitext(os.path.abspath(file))[0]
        with open(file, "w") as f:
            f.write(
                self.yaml(
                    base_key=base_key,
                    file_prefix=file_prefix,
                    save_models=save_models,
                    save_jit=save_jit,
                )
            )

    @classmethod
    def from_file(cls, filename: str):
        if not os.path.exists(filename):
            raise OSError(f"File {filename} is not found.")
        with open(filename, "r") as file:
            return cls.from_yaml(file)

    @classmethod
    def from_yaml(cls, yaml_obj: [str, TextIOWrapper]):
        return cls.model_validate(parse_config(yaml_obj, cls.model_fields))

    def register_to_mlflow(
        self,
        input_dict: dict[str, Union[float, torch.Tensor]],
        artifact_path: str,
        registered_model_name: str | None = None,
        tags: dict[str, Any] | None = None,
        version_tags: dict[str, Any] | None = None,
        alias: str | None = None,
        run_name: str | None = None,
        log_model_dump: bool = True,
        save_jit: bool = False,
        **kwargs,
    ):
        """
        Registers the model to MLflow if mlflow is installed. Each time this function is called, a new version
        of the model is created. The model is saved to the tracking server or local directory, depending on the
        MLFLOW_TRACKING_URI.

        If no tracking server is set up, data and artifacts are saved directly under your current directory. To set up
        a tracking server, set the environment variable MLFLOW_TRACKING_URI, e.g. a local port/path. See
        https://mlflow.org/docs/latest/getting-started/intro-quickstart/ for more info.

        Args:
            input_dict: Input dictionary to infer the model signature.
            artifact_path: Path to store the model in MLflow.
            registered_model_name: Name of the registered model in MLflow. Optional.
            tags: Tags to add to the MLflow model. Optional.
            version_tags: Tags to add to this MLflow model version. Optional.
            alias: Alias to add to this MLflow model version. Optional.
            run_name: Name of the MLflow run. Optional.
            log_model_dump: Whether to log the model dump files as artifacts. Optional.
            save_jit: Whether to save the model as TorchScript when calling model.dump, if log_model_dump=True. Optional.
            **kwargs: Additional arguments for mlflow.pyfunc.log_model.

        Returns:
            Model info metadata, mlflow.models.model.ModelInfo.
        """
        return register_model(
            self,
            input_dict,
            artifact_path,
            registered_model_name,
            tags,
            version_tags,
            alias,
            run_name,
            log_model_dump,
            save_jit,
            **kwargs,
        )
